/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.apphelpers;


/**
 * Fixed-size cache of objects that allows removing only the least recently used one.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
final class FixedSizeLRUCache {

    /**
     * Array of buckets containing entries with "similar" hash values.
     */
    private final CacheEntry[] buckets;

    /**
     * Special bucket containing unused {@link CacheEntry} instances.
     */
    private CacheEntry freeEntriesBucket;

    /**
     * Reference to the first (youngest) element in the access list (list of entries ordered by
     * access time) or <code>null</code> if the list is empty.
     */
    private CacheEntry accessListFirst;

    /**
     * Reference to the last (oldest) element in the access list (list of entries ordered by
     * access time) or <code>null</code> if the list is empty.
     */
    private CacheEntry accessListLast;

    /**
     * Key used as a parameter for the recent {@link #get(Object)} operation or
     * <code>null</code> if the operation has not been used yet.
     */
    private Object recentlyGetKey;

    /**
     * Entry returned as a result of the recent {@link #get(Object)} or <code>null</code> if the
     * operation has not been used yet.
     */
    private CacheEntry recentlyGetEntry;

    /**
     * Constructs <code>FixedSizeLRUCache</code> instance that can hold up to given number of
     * key-value pairs.
     *
     * @param size size of the cache (maximum number of entries if can hold)
     */
    FixedSizeLRUCache(int size) {
        if (size <= 0) {
            throw new IllegalArgumentException(
                "Cache size must be positive (given size: " + size + ")");
        }
        buckets = new CacheEntry[size * 2];
        for (int i = 0; i < size; i++) {
            addToFreeEntries(new CacheEntry());
        }
    }

    /**
     * Adds given key-value pair to the cache. Given key must not exist in the cache.
     *
     * @param key key used to identify the value
     * @param value the value or <code>null</code>
     * @throws DuplicateKeyException in case an entry with given key already exists in the cache
     * @throws SizeExceededException in case the cache is already full
     */
    public void add(Object key, Object value) {
        int keyHashCode = key.hashCode();
        int bucketIdx = getBucketIndex(keyHashCode);
        CacheEntry oldEntry;
        if (key.equals(recentlyGetKey)) {
            oldEntry = recentlyGetEntry;
        } else {
            oldEntry = getFromBucket(bucketIdx, key, keyHashCode);
        }
        if (oldEntry != null) {
            throw new DuplicateKeyException(key);
        }
        CacheEntry entry = getFreeEntry();
        entry.set(key, keyHashCode, value);
        addToBucketAsFirst(bucketIdx, entry);
        addToAccessListAsFirst(entry);
    }

    /**
     * Returns a value associated with given key (previously added to the cache using
     * {@link #add(Object, Object)} method).
     *
     * @param key key used to identify the value
     * @return
     *     value associated with given key or <code>null</code> if there's no entry with given
     *     key in the cache
     */
    public Object get(Object key) {
        int keyHashCode = key.hashCode();
        int bucketIdx = getBucketIndex(keyHashCode);
        CacheEntry foundEntry = getFromBucket(bucketIdx, key, keyHashCode);
        recentlyGetKey = key;
        recentlyGetEntry = foundEntry;
        if (foundEntry != null) {
            if (foundEntry.prevInBucket != null) {
                removeFromBucket(bucketIdx, foundEntry);
                addToBucketAsFirst(bucketIdx, foundEntry);
            }
            if (foundEntry.prevInAccessList != null) {
                removeFromAccessList(foundEntry);
                addToAccessListAsFirst(foundEntry);
            }
            return foundEntry.value;
        } else {
            return null;
        }
    }

    /**
     * Removes the least recently used entry from the cache.
     *
     * @return
     *     removed entry; it will contain removed key and value only until next modification
     *     of the cache, after that the object should not be used
     */
    public CacheEntry removeLeastRecentlyUsed() {
        CacheEntry entry = accessListLast;
        if (entry != null) {
            int bucketIdx = getBucketIndex(entry.key.hashCode());
            removeFromBucket(bucketIdx, entry);
            removeFromAccessList(entry);
            addToFreeEntries(entry);
        }
        return entry;
    }

    /**
     * Tells whether the cache has any free space for additional entries.
     *
     * @return
     *     <code>true</code> if at least one additional entry may be added to the cache,
     *     <code>false</code> otherwise
     */
    public boolean hasFreeSpace() {
        return freeEntriesBucket != null;
    }

    /**
     * Returns index of bucket that should be used for object with given hash code.
     *
     * @param hashCode hash code of an object
     * @return index of bucket
     */
    private int getBucketIndex(int hashCode) {
        return Math.abs(hashCode % buckets.length);
    }

    /**
     * Searches bucket with given index for an entry with given key.
     *
     * @param bucketIdx index of bucket to search in
     * @param key key to search for
     * @param keyHashCode hash code of the key, it must be equal to <code>key.hashCode()</code>
     * @return
     *     found entry with specified key or <code>null</code> if no such entry exists in the
     *     bucket
     */
    private CacheEntry getFromBucket(int bucketIdx, Object key, int keyHashCode) {
        CacheEntry foundEntry = null;
        for (CacheEntry e = buckets[bucketIdx]; e != null; e = e.nextInBucket) {
            if (e.keyHashCode == keyHashCode && e.key.equals(key)) {
                foundEntry = e;
                break;
            }
        }
        return foundEntry;
    }

    /**
     * Removes specified entry from bucket with given index.
     *
     * @param bucketIdx index of bucket to modify
     * @param entry entry to remove, it must be located in the bucket
     */
    private void removeFromBucket(int bucketIdx, CacheEntry entry) {
        if (entry.prevInBucket != null) {
            entry.prevInBucket.nextInBucket = entry.nextInBucket;
        }
        if (entry.nextInBucket != null) {
            entry.nextInBucket.prevInBucket = entry.prevInBucket;
        }
        if (buckets[bucketIdx] == entry) {
            buckets[bucketIdx] = entry.nextInBucket;
        }
    }

    /**
     * Adds specified entry to bucket with given index, locating it at the beginning of list of
     * entries.
     *
     * @param bucketIdx index of bucket to modify
     * @param entry entry to add to the bucket, it may not be located in any bucket
     */
    private void addToBucketAsFirst(int bucketIdx, CacheEntry entry) {
        entry.prevInBucket = null;
        entry.nextInBucket = buckets[bucketIdx];
        if (buckets[bucketIdx] != null) {
            buckets[bucketIdx].prevInBucket = entry;
        }
        buckets[bucketIdx] = entry;
    }

    /**
     * Removes given entry from the access list (list of entries ordered by access time).
     *
     * @param entry entry to remove from the list
     */
    private void removeFromAccessList(CacheEntry entry) {
        if (entry.prevInAccessList != null) {
            entry.prevInAccessList.nextInAccessList = entry.nextInAccessList;
        }
        if (entry.nextInAccessList != null) {
            entry.nextInAccessList.prevInAccessList = entry.prevInAccessList;
        }
        if (accessListFirst == entry) {
            accessListFirst = entry.nextInAccessList;
        }
        if (accessListLast == entry) {
            accessListLast = entry.prevInAccessList;
        }
    }

    /**
     * Adds given entry to the access list (list of entries ordered by access time), locating it
     * at the beginning of the list.
     *
     * @param entry entry to add to the list
     */
    private void addToAccessListAsFirst(CacheEntry entry) {
        entry.prevInAccessList = null;
        entry.nextInAccessList = accessListFirst;
        if (accessListFirst != null) {
            accessListFirst.prevInAccessList = entry;
        } else {
            accessListLast = entry;
        }
        accessListFirst = entry;
    }

    /**
     * Adds given entry to the list of free (unused) entries.
     *
     * @param entry entry to add to the list
     */
    private void addToFreeEntries(CacheEntry entry) {
        entry.prevInBucket = null;
        entry.nextInBucket = freeEntriesBucket;
        if (freeEntriesBucket != null) {
            freeEntriesBucket.prevInBucket = entry;
        }
        freeEntriesBucket = entry;
    }

    /**
     * Gets an entry from the list of free (unused) entries. The entry will be removed from the
     * list.
     *
     * @return entry obtained from the list
     * @throws SizeExceededException in case there're no more free entries
     */
    private CacheEntry getFreeEntry() {
        if (freeEntriesBucket == null) {
            throw new SizeExceededException();
        }
        CacheEntry entry = freeEntriesBucket;
        freeEntriesBucket = entry.nextInBucket;
        if (freeEntriesBucket != null) {
            freeEntriesBucket.prevInBucket = null;
        }
        return entry;
    }


    /**
     * Object describing single cache entry, i.e. single key-value pair.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    public static final class CacheEntry {

        /**
         * Key used to identify this entry.
         */
        private Object key;

        /**
         * Hash code of the key for faster lookup, always equal to <code>key.hashCode()</code>
         */
        private int keyHashCode;

        /**
         * The actual value stored in this entry.
         */
        private Object value;

        /**
         * Previous entry in bucket of this entry or <code>null</code> if this entry is the
         * first one in the list.
         */
        private CacheEntry prevInBucket;

        /**
         * Next entry in bucket of this entry or <code>null</code> if this entry is the last one
         * in the list.
         */
        private CacheEntry nextInBucket;

        /**
         * Previous entry in the access list or <code>null</code> if this entry is the first
         * one in the list.
         */
        private CacheEntry prevInAccessList;

        /**
         * Next entry in the access list or <code>null</code> if this entry is the last one in
         * the list.
         */
        private CacheEntry nextInAccessList;

        /**
         * The default constructor
         */
        private CacheEntry() { }

        /**
         * Sets new key-value pair for this entry.
         *
         * @param newKey key used to identify this entry and its value
         * @param newKeyHashCode
         *     hash code of the key, it must be equal to <code>key.hashCode()</code>
         * @param newValue value to store in this entry
         */
        private void set(Object newKey, int newKeyHashCode, Object newValue) {
            key = newKey;
            keyHashCode = newKeyHashCode;
            value = newValue;
        }

        /**
         * Returns key of this cache entry which is used to identify it in the cache.
         *
         * @return key of this entry
         */
        public Object getKey() {
            return key;
        }

        /**
         * Returns the actual value stored in this entry.
         *
         * @return value of this entry
         */
        public Object getValue() {
            return value;
        }

    }


    /**
     * Exception indicating an attempt to add an object to the cache using a key that already
     * exists.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    public static final class DuplicateKeyException extends RuntimeException {

        /**
         * Constructs <code>DuplicateKeyException</code> instance for given duplicate key.
         *
         * @param key duplicated key
         */
        private DuplicateKeyException(Object key) {
            super("Entry with key \"" + key + "\" already exists.");
        }

    }


    /**
     * Exception indicating an attempt to add an object to the cache while it's already full.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    public static final class SizeExceededException extends RuntimeException {

        /**
         * Constructs <code>SizeExceededException</code> instance.
         */
        private SizeExceededException() {
            super("The cache doesn't contain any free entries.");
        }

    }

}
