/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.apphelpers;

import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.parser.ExprTreePool;
import pl.chyla.wyjec.parser.ExpressionTree;
import pl.chyla.wyjec.parser.InvalidInputException;
import pl.chyla.wyjec.parser.Parser;
import pl.chyla.wyjec.stdconsts.StandardConstantsParser;
import pl.chyla.wyjec.stdfuncs.StandardFunctions;
import pl.chyla.wyjec.utils.RawString;


/**
 * Wrapper class for standard Wyjec {@link Parser parser}, offering simpler but less flexible
 * interface. It should be used only if avoiding memory allocations during parsing is not an
 * important requirement.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class SimpleParser {

    /**
     * Wyjec parser to use for real processing.
     */
    private final Parser parser;

    /**
     * Constructs <code>SimpleParser</code> instance using the default configuration and with
     * support for all standard Wyjec functions and constants.
     */
    public SimpleParser() {
        this(
            Config.getDefaultConfig(), StandardFunctions.getFunctions(),
            StandardConstantsParser.getInstance());
    }

    /**
     * Constructs <code>SimpleParser</code> instance using given configuration, functions and
     * constants parser.
     *
     * @param config configuration for the parser
     * @param funcs array of functions to use during semantic analysis of expressions
     * @param constsParser constants parser to use during semantic analysis of expressions
     */
    public SimpleParser(Config config, Function[] funcs, IConstantsParser constsParser) {
        parser = new Parser(config, funcs, constsParser);
    }

    /**
     * Parses given Wyjec expression and generates the corresponding expression tree. The
     * generated tree will consist of newly created objects and will be fully independent from
     * this parser and from other expression trees.
     * <p>
     * Note: this method may create a lot of objects; if you want to avoid this, consider
     * accessing {@link Parser} class directly.
     *
     * @param expression string representation of an expression to parse
     * @return expression tree representing parsed expression
     * @throws InvalidInputException in case of syntax or semantic error in the given expression
     */
    public ExpressionTree parse(String expression) throws InvalidInputException {
        ExpressionTree exprTree = parser.parse(new RawString(expression));
        ExprTreePool privPool = new ExprTreePool(1, Config.getZeroMemLimitsConfig());
        return exprTree.makeDeepCopy(privPool);
    }

}
