/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.core;


/**
 * Class describing a function's signature, i.e. types of its parameters (including optional
 * ones).
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FunctionSignature {

    /**
     * <code>FunctionSignature</code> instance representing a signature of a function that has
     * no parameters. It may be used as a value returned by {@link Function#getSignature()}
     * inside implementations of parameter-less functions.
     */
    public static final FunctionSignature SIGNATURE_NO_PARAMETERS =
        new FunctionSignature(new Parameter[0], null);

    /**
     * Expected types of required function's parameters
     */
    private final Type[] reqParamsTypes;

    /**
     * Expected type of optional function's parameters
     */
    private final Type optParamsType;

    /**
     * Informs whether actual parameters matching parameters declared by this object may be
     * evaluated lazily.
     */
    private final boolean lazyEvaluationSupport;

    /**
     * Constructs new function's signature corresponding to a function with variable or constant
     * number of accepted parameters. If <code>optParams</code> is <code>null</code>, the
     * function accepts only constant number of required parameters, otherwise it accepts
     * constant number required parameters + zero or more optional parameters. Additionally,
     * this constructor allows to specify using <code>lazyEvaluation</code> parameter, whether
     * function parameters declared with this object may be evaluated lazily.
     *
     * @param params
     *     objects representing function's declared parameters in the appropriate order,
     *     describing their types
     * @param optParams
     *     object representing a list of optional function's parameters that may be added
     *     immediately after the required ones (all optional parameters have the same type); if
     *     the parameter is <code>null</code> then the function doesn't accept any optional
     *     parameters
     * @param lazyEvaluation
     *     defines whether function parameters declared with this object may be evaluated lazily
     *     during evaluation of function that uses them; <code>true</code> means that lazy
     *     evaluation is supported, <code>false</code> - that it's not
     */
    public FunctionSignature(
        Parameter[] params, OptParameters optParams, boolean lazyEvaluation) {
        for (int i = 0; i < params.length; i++) {
            params[i].configure(i);
        }
        if (optParams != null) {
            optParams.configure(params.length);
        }
        this.reqParamsTypes = new Type[params.length];
        for (int i = 0; i < params.length; i++) {
            this.reqParamsTypes[i] = params[i].getType();
        }
        if (optParams != null) {
            this.optParamsType = optParams.getType();
        } else {
            this.optParamsType = Type.NONE;
        }
        this.lazyEvaluationSupport = lazyEvaluation;
    }

    /**
     * Constructs new function's signature corresponding to a function with variable or constant
     * number of accepted parameters. If <code>optParams</code> is <code>null</code>, the
     * function accepts only constant number of required parameters, otherwise it accepts
     * constant number required parameters + zero or more optional parameters. When evaluating
     * functions corresponding to this signature, all actual parameters will have to be
     * evaluated earlier, i.e. without using lazy evaluation.
     *
     * @param params
     *     objects representing function's declared parameters in the appropriate order,
     *     describing their types
     * @param optParams
     *     object representing a list of optional function's parameters that may be added
     *     immediately after the required ones (all optional parameters have the same type); if
     *     the parameter is <code>null</code> then the function doesn't accept any optional
     *     parameters
     */
    public FunctionSignature(Parameter[] params, OptParameters optParams) {
        this(params, optParams, false);
    }

    /**
     * Returns array of types for declared required parameters of a function connected with the
     * signature. Contents of returned array may not be modified. Returned value is
     * pre-calculated, not generated on demand. It should not be modified.
     *
     * @return array of types of required parameters of a function connected with the signature
     */
    public Type[] getRequiredParametersTypes() {
        return reqParamsTypes;
    }

    /**
     * Returns type of declared optional parameters of a function connected with the signature.
     * All optional parameters must have the same type so the returned value describes all of
     * them. If the function doesn't accept optional parameters, this function returns
     * {@link Type#NONE}.
     *
     * @return type of optional parameters of a function connected with the signature
     */
    public Type getOptionalParametersType() {
        return optParamsType;
    }

    /**
     * Informs whether actual parameters matching parameters declared by this object may be
     * evaluated lazily, i.e. on demand during function evaluation. Without lazy evaluation
     * support, all parameters should be evaluated when function evaluation starts.
     *
     * @return
     *     <code>true</code> if actual parameters of this function may be evaluated lazily and
     *     <code>false</code> otherwise.
     */
    public boolean getLazyEvaluationSupport() {
        return lazyEvaluationSupport;
    }

    /**
     * Returns a string representing types of parameters accepted by a function connected with
     * the signature, enclosed in a pair of braces. If the function accepts no parameters,
     * <code>"()"</code> is returned. Required parameters are represented by names of their
     * types, optional parameters are represented by single name of a type followed by dots
     * ("...").
     *
     * @return
     *     a string representing types of declared parameters of a function connected with the
     *     signature
     */
    public String getParametersTypesDescription() {
        StringBuffer desc = new StringBuffer();
        desc.append('(');
        if (reqParamsTypes.length != 0) {
            desc.append(reqParamsTypes[0].getName());
            for (int i = 1; i < reqParamsTypes.length; i++) {
                desc.append(',');
                desc.append(reqParamsTypes[i].getName());
            }
        }
        if (optParamsType != Type.NONE) {
            if (reqParamsTypes.length != 0) {
                desc.append(',');
            }
            desc.append(optParamsType.getName());
            desc.append("...");
        }
        desc.append(')');
        return desc.toString();
    }

}
