/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import pl.chyla.wyjec.core.Type;


/**
 * Base class for classes representing "atom" node in expression tree (the equivalent of "atom"
 * nonterminal in Wyjec grammar).
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class AtomNode {

    /**
     * Returns the actual type of this atom, which is needed to check which <code>getAs*</code>
     * method may be called on it.
     *
     * @return the type of this atom
     */
    public abstract AtomType getAtomType();

    /**
     * Returns this node using {@link ValueNode} as its static type. This method may be called
     * only if {@link #getAtomType()} method returns {@link AtomType#VALUE}.
     *
     * @return this node as {@link ValueNode} reference
     * @throws UnsupportedOperationException
     *     in case this node doesn't represent constant value, i.e. {@link #getAtomType()}
     *     method doesn't return {@link AtomType#VALUE}
     */
    public abstract ValueNode getAsValueNode();

    /**
     * Returns this node using {@link FunctionNode} as its static type. This method may be
     * called only if {@link #getAtomType()} method returns {@link AtomType#FUNCTION}.
     *
     * @return this node as {@link FunctionNode} reference
     * @throws UnsupportedOperationException
     *     in case this node doesn't represent reference to function i.e. {@link #getAtomType()}
     *     method doesn't return {@link AtomType#FUNCTION}
     */
    public abstract FunctionNode getAsFunctionNode();

    /**
     * Returns type of values that this atom may evaluate to. In case of atom representing
     * constant value it's a type of the value, in case of atom representing a function, it's
     * result type of the function.
     *
     * @return type of values evaluated from this atom
     */
    public abstract Type getValueType();

    /**
     * Makes deep copy of part of expression tree for which this atom is top-level node, i.e.
     * copies this atom and all nodes referenced from it. Given pool will be used to get
     * instances for copies of this atom and all referenced nodes of expression tree.
     *
     * @param pool
     *     pool used as the source of instances that will represent nodes of the copied part of
     *     expression tree
     * @return
     *     <code>AtomNode</code> instance taken from given pool and representing copy of part
     *     of expression tree starting from this atom
     */
    abstract AtomNode makeDeepCopy(ExprTreePool pool);

    /**
     * Recursively returns this node and all its inferior elements of expression tree to given
     * pool. This instance should be one of instances previously acquired from the same pool.
     * After calling this method this object should no longed be used.
     *
     * @param pool
     *     pool of expression tree objects that should be used to release this node and its
     *     children
     */
    abstract void deepRelease(ExprTreePool pool);

}
