/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.Type;


/**
 * Expression tree node representing reference to a function and its parameters that should be
 * used during evaluation.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FunctionNode extends AtomNode {

    /**
     * Function that this node represents reference to
     */
    private Function function;

    /**
     * First parameter of the function (other parameters can be accessed from it)
     */
    private ParameterNode parameters;

    /**
     * Constructs <code>FunctionNode</code> instance.
     */
    FunctionNode() { }

    /**
     * {@inheritDoc}
     *
     * @return {@link AtomType#FUNCTION}
     */
    public AtomType getAtomType() {
        return AtomType.FUNCTION;
    }

    /**
     * Always throws {@link UnsupportedOperationException} as instances of this class do not
     * represent constant values.
     *
     * @return nothing as the function never returns normally
     */
    public ValueNode getAsValueNode() {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns this node using {@link FunctionNode} as its static type.
     *
     * @return this instance using {@link FunctionNode} as its static type
     */
    public FunctionNode getAsFunctionNode() {
        return this;
    }

    /**
     * Initializes this <code>FunctionAtom</code> instance, setting its internal field.
     *
     * @param newFunction function to be referenced by this atom
     * @param newParameters
     *     parameters of the function to be referenced by this atom or <code>null</code> if
     *     there're no parameters
     */
    void set(Function newFunction, ParameterNode newParameters) {
        function = newFunction;
        parameters = newParameters;
    }

    /**
     * Returns return type of function that this node represents reference to.
     *
     * @return return type of function referenced by this node
     */
    public Type getValueType() {
        return function.getResultType();
    }

    /**
     * Returns function that this node refers to.
     *
     * @return function referenced by this node
     */
    public Function getFunction() {
        return function;
    }

    /**
     * Returns part of expression tree representing parameters for function referenced by this
     * node. Returned instance represents first parameter of the function and others can be
     * accessed from it (all parameters form linked list of nodes).
     *
     * @return node representing first parameter of function referenced by this node
     */
    public ParameterNode getParameters() {
        return parameters;
    }

    /**
     * {@inheritDoc}
     *
     * @param pool {@inheritDoc}
     * @return {@inheritDoc}
     */
    AtomNode makeDeepCopy(ExprTreePool pool) {
        FunctionNode copy = pool.getFunctionNode();
        copy.function = function;
        if (parameters != null) {
            copy.parameters = parameters.makeDeepCopy(pool);
        } else {
            copy.parameters = null;
        }
        return copy;
    }

    /**
     * Recursively returns this node and all its inferior elements of expression tree to given
     * pool. This instance should be one of <code>FunctionNode</code> instances previously
     * acquired with {@link ExprTreePool#getFunctionNode()} method from the same pool. After
     * calling this method this object should no longed be used.
     *
     * @param pool {@inheritDoc}
     */
    void deepRelease(ExprTreePool pool) {
        if (parameters != null) {
            parameters.deepRelease(pool);
        }
        pool.releaseFunctionNode(this);
    }

}
