/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.core.Value;


/**
 * Expression tree node representing constant value. It has no direct equivalent in Wyjec
 * grammar, because constant value may be generated both from <code>STRING</code> terminal and
 * from some parameterless functions implemented as "constants". In the latter case this node
 * may represent value of any Wyjec type.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class ValueNode extends AtomNode {

    /**
     * Value contained in this node
     */
    private Value value;

    /**
     * Constructs <code>ValueNode</code> instance.
     */
    ValueNode() { }

    /**
     * {@inheritDoc}
     *
     * @return {@link AtomType#VALUE}
     */
    public AtomType getAtomType() {
        return AtomType.VALUE;
    }

    /**
     * Returns this node using {@link ValueNode} as its static type.
     *
     * @return this instance using {@link ValueNode} as its static type
     */
    public ValueNode getAsValueNode() {
        return this;
    }

    /**
     * Always throws {@link UnsupportedOperationException} as instances of this class do not
     * represent references to functions.
     *
     * @return nothing as the function never returns normally
     */
    public FunctionNode getAsFunctionNode() {
        throw new UnsupportedOperationException();
    }

    /**
     * Initializes this <code>ValueAtom</code> instance, setting its internal field.
     *
     * @param newValue constant value to be represented by this atom
     */
    void set(Value newValue) {
        value = newValue;
    }

    /**
     * Returns type of constant value represented by this node.
     *
     * @return type of this node's value
     */
    public Type getValueType() {
        return value.getType();
    }

    /**
     * Returns constant value represented by this atom.
     *
     * @return this node's value
     */
    public Value getValue() {
        return value;
    }

    /**
     * {@inheritDoc}
     *
     * @param pool {@inheritDoc}
     * @return {@inheritDoc}
     */
    AtomNode makeDeepCopy(ExprTreePool pool) {
        ValueNode copy = pool.getValueNode();
        copy.value = pool.getValue();
        copy.value.set(value);
        return copy;
    }

    /**
     * Recursively returns this node and all its inferior elements of expression tree to given
     * pool. This instance should be one of <code>ValueNode</code> instances previously acquired
     * with {@link ExprTreePool#getValueNode()} method from the same pool. After calling this
     * method this object should no longed be used.
     *
     * @param pool {@inheritDoc}
     */
    void deepRelease(ExprTreePool pool) {
        pool.releaseValue(value);
        pool.releaseValueNode(this);
    }

}
