/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdconsts;

import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.CharUtils;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Parser for integer constants from range [-999999999999999999, 999999999999999999]
 * (from -10<sup>18</sup> - 1 to 10<sup>18</sup> - 1). This parser recognizes all string
 * representations of integer constants from this range, matching regular expression
 * "[+-]?[0-9]+", e.g. <code>123</code>, <code>0007</code>, <code>9</code>, <code>-78</code>,
 * <code>-0</code>.
 *
 * @see IConstantsParser
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class IntegerConstantsParser implements IConstantsParser {

    /**
     * Number of decimal digits in the maximum number recognized by the parser
     */
    private static final int MAX_NUMBER_DIGITS_COUNT = 18;

    /**
     * Instance of <code>IntegerConstantsParser</code>, accessible from outside via
     * {@link #getInstance()}.
     */
    private static final IntegerConstantsParser instance = new IntegerConstantsParser();

    /**
     * Private (empty) constructor to disallow instantiating of the class from outside. Class'
     * instance should be accessed via {@link #getInstance()} getter.
     */
    private IntegerConstantsParser() { }

    /**
     * Returns instance of <code>IntegerConstantsParser</code> which can be safely used
     * independly from other threads which also get <code>IntegerConstantsParser</code> using
     * this method. It's undefined whether this method returns newly created instance or not.
     *
     * @return instance of <code>IntegerConstantsParser</code>
     */
    public static IntegerConstantsParser getInstance() {
        return instance;
    }

    /**
     * Parses given string as a representation of integer constant value. If the string contains
     * decimal representation of integer from range [0, 999999999999999999], its integer value
     * will be stored in given {@link Value} and can be later accessed using
     * {@link Value#getInteger()} method.
     *
     * @param result {@link Value} object used to store recognized integer value
     * @param constStr string representation of constant to parse
     * @return
     *     <code>true</code> if given string contains decimal representation of integer from
     *     range [0, 999999999999999999] (in this case its value can be accessed using given
     *     {@link Value} object), <code>false</code> otherwise.
     * @see IConstantsParser#parseConstant(Value, ReadableRawString)
     */
    public boolean parseConstant(Value result, ReadableRawString constStr) {
        char[] chars = constStr.getStorage();
        int startPos = constStr.getStartPos();
        int endPos = constStr.getEndPos();
        if (!isValidInteger(chars, startPos, endPos)) {
            return false;
        }
        int curPos = startPos;
        long sign;
        if (chars[curPos] == '+') {
            sign = 1L;
            curPos++;
        } else if (chars[curPos] == '-') {
            sign = -1L;
            curPos++;
        } else {
            sign = 1L;
        }
        long value = 0L;
        while (curPos < endPos && chars[curPos] == '0') {
            curPos++;
        }
        while (curPos < endPos) {
            value = value * 10L + (chars[curPos] - '0');
            curPos++;
        }
        result.setInteger(sign * value);
        return true;
    }

    /**
     * Checks whether given sequence of characters (specified as part of given array) contains
     * valid integer constant. Valid float constant matches the following regular expression:
     * "[+-]?[0-9]+" and has up to {@link #MAX_NUMBER_DIGITS_COUNT} digits after leading zeros.
     *
     * @param chars array containing sequence of characters to check
     * @param startPos position of first character to check in the array
     * @param endPos position after the last character to check in the array
     * @return
     *     <code>true</code> if specified sequence of characters contains valid integer
     *     constant, <code>false</code> otherwise.
     */
    private boolean isValidInteger(char[] chars, int startPos, int endPos) {
        int curPos = startPos;
        if (curPos < endPos && (chars[curPos] == '+' || chars[curPos] == '-')) {
            curPos++;
        }
        int firstDigitPos = curPos;
        while (curPos < endPos && chars[curPos] == '0') {
            curPos++;
        }
        boolean isValid;
        if (firstDigitPos < endPos && endPos - curPos <= MAX_NUMBER_DIGITS_COUNT) {
            while (curPos < endPos && CharUtils.isDigit(chars[curPos])) {
                curPos++;
            }
            isValid = curPos == endPos;
        } else {
            isValid = false;
        }
        return isValid;
    }

}
