/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdconsts;

import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Parser for constant values that delegates parsing requests to specified group of other
 * parsers.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class StackedConstantsParser implements IConstantsParser {

    /**
     * Parsers that all parsing requests are delegated to
     */
    private final IConstantsParser[] parsers;

    /**
     * Constructs new parser based on given group of other parsers forming a stack.
     *
     * @param parsers
     *     array of parsers representing a stack, where elements with smaller indices are
     *     placed closer to the top of the stack
     */
    public StackedConstantsParser(IConstantsParser[] parsers) {
        this.parsers = parsers;
    }

    /**
     * Parses given string as a representation of constant value by relaying parsing request
     * to parsers from the underlying stack. Relaying begins from the top of the stack and goes
     * down to the bottom or until some parser parses given string successfully.
     *
     * @param result {@link Value} object used to store recognized constant value
     * @param constStr string representation of constant to parse
     * @return
     *     <code>true</code> if one of parsers from the stack successfully parses given string
     *     as constant (in this case its value can be accessed using given {@link Value}
     *     object), <code>false</code> otherwise.
     */
    public boolean parseConstant(Value result, ReadableRawString constStr) {
        for (int i = 0; i < parsers.length; i++) {
            if (parsers[i].parseConstant(result, constStr)) {
                return true;
            }
        }
        return false;
    }

}
