/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.datefuncs.basic;

import java.util.Arrays;
import java.util.Calendar;
import pl.chyla.wyjec.utils.LongFormatter;
import pl.chyla.wyjec.utils.RawString;


/**
 * Helper class grouping operations for formatting dates.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
final class DateFormatUtils {

    /**
     * Number of characters occupied by a representation of <code>Calendar.YEAR</code> field.
     */
    private static final int YEAR_CHARS_COUNT = 4;

    /**
     * Number of characters occupied by a representation of <code>Calendar.MONTH</code> field.
     */
    private static final int MONTH_CHARS_COUNT = 2;

    /**
     * Number of characters occupied by a representation of <code>Calendar.DAY_OF_MONTH</code>
     * field.
     */
    private static final int DAY_CHARS_COUNT = 2;

    /**
     * Number of characters occupied by a representation of <code>Calendar.HOUR_OF_DAY</code>
     * field.
     */
    private static final int HOUR_CHARS_COUNT = 2;

    /**
     * Number of characters occupied by a representation of <code>Calendar.MINUTE</code> field.
     */
    private static final int MINUTE_CHARS_COUNT = 2;

    /**
     * Number of characters occupied by a representation of <code>Calendar.SECOND</code> field.
     */
    private static final int SECOND_CHARS_COUNT = 2;

    /**
     * Private empty constructor to disallow instantiating of the class.
     */
    private DateFormatUtils() { }

    /**
     * Appends ISO representation (YYYY-mm-dd) of the date stored in the calendar to the result.
     *
     * @param result string to append the result to
     * @param cal calendar containing date to use
     */
    static void appendISODate(RawString result, Calendar cal) {
        appendDigits(result, cal.get(Calendar.YEAR), YEAR_CHARS_COUNT);
        result.append('-');
        appendDigits(result, cal.get(Calendar.MONTH) + 1, MONTH_CHARS_COUNT);
        result.append('-');
        appendDigits(result, cal.get(Calendar.DAY_OF_MONTH), DAY_CHARS_COUNT);
    }

    /**
     * Appends ISO representation (HH:MM:SS) of the time of the day stored in the calendar to
     * the result.
     *
     * @param result string to append the result to
     * @param cal calendar containing date to use
     */
    static void appendISOTime(RawString result, Calendar cal) {
        appendDigits(result, cal.get(Calendar.HOUR_OF_DAY), HOUR_CHARS_COUNT);
        result.append(':');
        appendDigits(result, cal.get(Calendar.MINUTE), MINUTE_CHARS_COUNT);
        result.append(':');
        appendDigits(result, cal.get(Calendar.SECOND), SECOND_CHARS_COUNT);
    }

    /**
     * Appends string representation of given number to given {@link RawString} instance, using
     * at least specified number of digits. In case of missing digits the appropriate number of
     * zeros will be prepended.
     *
     * @param result string to append the result to
     * @param value numeric value to append to the result
     * @param minDigitsCount minimum number of digits to append
     */
    static void appendDigits(RawString result, int value, int minDigitsCount) {
        char[] storage = result.getStorage();
        int digitsCount = LongFormatter.getCharactersCount(value);
        if (digitsCount < minDigitsCount) {
            int prefixLength = minDigitsCount - digitsCount;
            int prevEndPos = result.getEndPos();
            result.extend(prefixLength);
            Arrays.fill(storage, prevEndPos, prevEndPos + prefixLength, '0');
        }
        int prevEndPos = result.getEndPos();
        result.extend(digitsCount);
        LongFormatter.formatToArray(value, storage, prevEndPos);
    }

}
