/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.datefuncs.basic;

import java.util.Calendar;
import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.StringFunction;
import pl.chyla.wyjec.core.parameters.DateParameter;
import pl.chyla.wyjec.core.parameters.StringParameter;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;
import pl.chyla.wyjec.utils.SimpleCalendar;


/**
 * Returns string representation of given date generated according to given format
 * specification. Special sequences of characters contained in format specification will be
 * replaced with the appropriate attributes of the date, all other characters will be copied
 * without changes. Special sequence consists of "%" character, followed by one of the
 * following:
 * <ul>
 * <li>Y: 4-digit number of the year</li>
 * <li>y: two last digits of number of the year (year % 100)</li>
 * <li>m: 2-digit number of the month (range: 1 to 12)</li>
 * <li>d: 2-digit number of day of the month (range: 1 to 31)</li>
 * <li>H: 2-digit number of the hour using 24-hour clock (range: 0 to 23)</li>
 * <li>I: 2-digit number of the hour using 12-hour clock (range: 0 to 11)</li>
 * <li>M: 2-digit number of the minute (range: 0 to 59)</li>
 * <li>S: 2-digit number of the second (range: 0 to 59)</li>
 * <li>T: time of the day given in format specified by ISO 8601 standard
 * (HH:MM:SS)</li>
 * <li>D: date given in format specified by ISO 8601 standard (YYYY-mm-dd)</li>
 * <li>%: % character</li>
 * </ul>
 * Sequence consisting of "%" character followed by character not on the list above or single
 * "%" at the end of format string will be left unchanged.
 *
 * @wyjec.func format
 * @wyjec.example "{format|year: %Y, time: %T, 100%|{&#64;1977-03-09 21:00:00}}"
 * "year: 1977, time: 21:00:00, 100%" string
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FnFormat extends StringFunction {

    /**
     * Format specification for the resulting string
     * @wyjec.param format
     */
    private static final StringParameter pFormat = new StringParameter();

    /**
     * Date to represent as a string
     * @wyjec.param date
     */
    private static final DateParameter pDate = new DateParameter();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pFormat, pDate}, null);

    /**
     * Character preceding every conversion character in the format string
     */
    private static final char CONV_PREFIX = '%';

    /**
     * Conversion character for 4-digit number of the year
     */
    private static final char YEAR_DDDD_CONV = 'Y';

    /**
     * Number of characters occupied by field specified by {@link #YEAR_DDDD_CONV}
     */
    private static final int YEAR_DDDD_CHARS_COUNT = 4;

    /**
     * Conversion character for two last digits of number of the year (year % 100)
     */
    private static final char YEAR_DD_CONV = 'y';

    /**
     * Number of characters occupied by field specified by {@link #YEAR_DD_CONV}
     */
    private static final int YEAR_DD_CHARS_COUNT = 2;

    /**
     * Conversion character for 2-digit number of the month (range: 1 to 12)
     */
    private static final char MONTH_DD_CONV = 'm';

    /**
     * Number of characters occupied by field specified by {@link #MONTH_DD_CONV}
     */
    private static final int MONTH_DD_CHARS_COUNT = 2;

    /**
     * Conversion character for 2-digit number of day of the month (range: 1 to 31)
     */
    private static final char DAY_DD_CONV = 'd';

    /**
     * Number of characters occupied by field specified by {@link #DAY_DD_CONV}
     */
    private static final int DAY_DD_CHARS_COUNT = 2;

    /**
     * Conversion character for 2-digit number of the hour using 24-hour clock (range: 0 to 23)
     */
    private static final char HOUR_24_DD_CONV = 'H';

    /**
     * Number of characters occupied by field specified by {@link #HOUR_24_DD_CONV}
     */
    private static final int HOUR_24_DD_CHARS_COUNT = 2;

    /**
     * Conversion character for 2-digit number of the hour using 12-hour clock (range: 0 to 11)
     */
    private static final char HOUR_12_DD_CONV = 'I';

    /**
     * Number of characters occupied by field specified by {@link #HOUR_12_DD_CONV}
     */
    private static final int HOUR_12_DD_CHARS_COUNT = 2;

    /**
     * Conversion character for 2-digit number of the minute (range: 0 to 59)
     */
    private static final char MINUTE_DD_CONV = 'M';

    /**
     * Number of characters occupied by field specified by {@link #MINUTE_DD_CONV}
     */
    private static final int MINUTE_DD_CHARS_COUNT = 2;

    /**
     * Conversion character for 2-digit number of the second (range: 0 to 59)
     */
    private static final char SECOND_DD_CONV = 'S';

    /**
     * Number of characters occupied by field specified by {@link #SECOND_DD_CONV}
     */
    private static final int SECOND_DD_CHARS_COUNT = 2;

    /**
     * Conversion character for time of the day given in format specified by ISO 8601 standard
     * (HH:MM:SS).
     */
    private static final char TIME_ISO_CONV = 'T';

    /**
     * Conversion character for date given in format specified by ISO 8601 standard (YYYY-mm-dd)
     */
    private static final char DATE_ISO_CONV = 'D';

    /**
     * Temporary calendar instance used for accessing date fields
     */
    private SimpleCalendar tmpCal = new SimpleCalendar();

    public FunctionSignature getSignature() {
        return signature;
    }

    public String getName() {
        return "format";
    }

    protected void execute(RawString result) {
        tmpCal.setTimeInMillis(getParameter(pDate));
        ReadableRawString fmt = getParameter(pFormat);
        char[] fmtChars = fmt.getStorage();
        int fmtEndPos = fmt.getEndPos();
        int curPos = fmt.getStartPos();
        while (curPos < fmtEndPos) {
            int oldPos = curPos;
            while (curPos < fmtEndPos && fmtChars[curPos] != CONV_PREFIX) {
                curPos++;
            }
            result.append(fmtChars, oldPos, curPos - oldPos);
            if (curPos < fmtEndPos) {
                curPos++;
                if (curPos < fmtEndPos) {
                    performConversion(result, fmtChars[curPos]);
                    curPos++;
                } else {
                    result.append(CONV_PREFIX);
                }
            }
        }
    }

    /**
     * Processes given conversion character by appending the appropriate string based on the
     * date stored in the temporary calendar to the result.
     *
     * @param convChar
     *     conversion character (character following "{@value #CONV_PREFIX}" in the format
     *     string
     */
    private void performConversion(RawString result, char convChar) {
        switch (convChar) {
        case YEAR_DDDD_CONV:
            appendFieldDigits(result, Calendar.YEAR, YEAR_DDDD_CHARS_COUNT);
            break;
        case YEAR_DD_CONV:
            DateFormatUtils.appendDigits(
                result, tmpCal.get(Calendar.YEAR) % 100, YEAR_DD_CHARS_COUNT);
            break;
        case MONTH_DD_CONV:
            DateFormatUtils.appendDigits(
                result, tmpCal.get(Calendar.MONTH) + 1, MONTH_DD_CHARS_COUNT);
            break;
        case DAY_DD_CONV:
            appendFieldDigits(result, Calendar.DAY_OF_MONTH, DAY_DD_CHARS_COUNT);
            break;
        case HOUR_24_DD_CONV:
            appendFieldDigits(result, Calendar.HOUR_OF_DAY, HOUR_24_DD_CHARS_COUNT);
            break;
        case HOUR_12_DD_CONV:
            appendFieldDigits(result, Calendar.HOUR, HOUR_12_DD_CHARS_COUNT);
            break;
        case MINUTE_DD_CONV:
            appendFieldDigits(result, Calendar.MINUTE, MINUTE_DD_CHARS_COUNT);
            break;
        case SECOND_DD_CONV:
            appendFieldDigits(result, Calendar.SECOND, SECOND_DD_CHARS_COUNT);
            break;
        case DATE_ISO_CONV:
            DateFormatUtils.appendISODate(result, tmpCal);
            break;
        case TIME_ISO_CONV:
            DateFormatUtils.appendISOTime(result, tmpCal);
            break;
        case CONV_PREFIX:
            result.append(CONV_PREFIX);
            break;
        default:
            result.append(CONV_PREFIX);
            result.append(convChar);
        }
    }

    /**
     * Appends current value of specified field from {@link #tmpCal} calendar to given
     * {@link RawString} instance, using at least specified number of digits.
     *
     * @param output string to append the result to
     * @param fieldNumber number specifying calendar field to use (one of Calendar.* constants)
     * @param minDigitsCount minimum number of digits to append
     */
    private void appendFieldDigits(RawString output, int fieldNumber, int minDigitsCount) {
        int fieldValue = tmpCal.get(fieldNumber);
        DateFormatUtils.appendDigits(output, fieldValue, minDigitsCount);
    }

}
