/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.io.PrintStream;
import java.io.PrintWriter;


/**
 * Subclass of <code>java.lang.RuntimeException</code> that supports exception chaining. The
 * class is provided for compatibility with Java 1.2 and it mimics the behaviour of
 * <code>RuntimeException</code> class in Java 1.4 (the most important feature is printing
 * stack traces for chained exceptions in <code>printStackTrace</code> methods).
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public class ChainedException extends RuntimeException {

    /**
     * The cause of this exception, set by constructor or {@link #initCause(Throwable)} method.
     */
    private Throwable exceptionCause;

    /**
     * Flag set to true every time {@link #getCause()} method is called. It's used by
     * <code>printStackTrace</code> methods to detect that <code>getCause()</code> has been used
     * by <code>printStackTrace</code> implementation in <code>java.lang.Exception</code> class.
     */
    private boolean causeReadFlag;

    /**
     * Constructs <code>ChainedException</code> instance without detail message attached (it
     * will be set to <code>null</code>).
     */
    public ChainedException() {
        super();
        exceptionCause = null;
    }

    /**
     * Constructs <code>ChainedException</code> instance with given detail message attached.
     *
     * @param msg the detail message (may be set to <code>null</code>)
     */
    public ChainedException(String msg) {
        super(msg);
        exceptionCause = null;
    }

    /**
     * Constructs <code>ChainedException</code> instance with given <code>Throwable</code> as a
     * cause.
     *
     * @param cause the cause (may be set to <code>null</code>)
     */
    public ChainedException(Throwable cause) {
        super();
        this.exceptionCause = cause;
    }

    /**
     * Constructs <code>ChainedException</code> instance with given detail message attached and
     * given <code>Throwable</code> as a cause.
     *
     * @param msg the detail message (may be set to <code>null</code>)
     * @param cause the cause (may be set to <code>null</code>)
     */
    public ChainedException(String msg, Throwable cause) {
        super(msg);
        this.exceptionCause = cause;
    }

    /**
     * Sets cause of this exception to given <code>Throwable</code>.
     *
     * @param cause <code>Throwable</code> to set as cause of this exception
     * @return this exception
     */
    public final Throwable initCause(Throwable cause) {
        if (cause == this) {
            throw new IllegalArgumentException(
                "ChainedException instance may not be used as a cause for itselt");
        }
        this.exceptionCause = cause;
        return this;
    }

    /**
     * Returns the cause of this exception, as given in
     * {@link #ChainedException(String, Throwable)} or
     * {@link #ChainedException(Throwable)} constructors or {@link #initCause(Throwable)}
     * or <code>null</code> value if the cause was not set.
     *
     * @return the cause of this exception
     */
    public final Throwable getCause() {
        causeReadFlag = true;
        return exceptionCause;
    }

    /**
     * Prints this exception, its stack trace and its cause to the standard error stream.
     */
    public final void printStackTrace() {
        printStackTrace(System.err);
    }

    /**
     * Prints this exception, its stack trace and its cause to the specified print stream.
     *
     * @param s print stream to write to
     */
    public final void printStackTrace(PrintStream s) {
        synchronized (this) {
            causeReadFlag = false;
            super.printStackTrace(s);
            if (!causeReadFlag && exceptionCause != null) {
                s.print("Caused by: ");
                exceptionCause.printStackTrace(s);
            }
        }
    }

    /**
     * Prints this exception, its stack trace and its cause to the specified print writer.
     *
     * @param s print writer to write to
     */
    public final void printStackTrace(PrintWriter s) {
        synchronized (this) {
            causeReadFlag = false;
            super.printStackTrace(s);
            if (!causeReadFlag && exceptionCause != null) {
                s.print("Caused by: ");
                exceptionCause.printStackTrace(s);
            }
        }
    }

}
